@@ -5,7 +5,7 @@
 # ColorDiff - a wrapper/replacment for 'diff' producing                #
 #             colourful output                                         #
 #                                                                      #
-# Copyright (C)2002-2007 Dave Ewart (davee@sungate.co.uk)              #
+# Copyright (C)2002-2009 Dave Ewart (davee@sungate.co.uk)              #
 #                                                                      #
 ########################################################################
 #                                                                      #
@@ -26,11 +26,11 @@ use Getopt::Long qw(:config pass_through);
 use IPC::Open2;
 
 my $app_name     = 'colordiff';
-my $version      = '1.0.7';
+my $version      = '1.0.9';
 my $author       = 'Dave Ewart';
 my $author_email = 'davee@sungate.co.uk';
 my $app_www      = 'http://colordiff.sourceforge.net/';
-my $copyright    = '(C)2002-2007';
+my $copyright    = '(C)2002-2009';
 my $show_banner  = 1;
 my $color_patch  = 0;
 
@@ -70,8 +70,10 @@ push (@config_files, "$ENV{HOME}/.colordiffrc") if (defined $ENV{HOME});
 my $config_file;
 
 foreach $config_file (@config_files) {
-    if (open(COLORDIFFRC, "<$config_file")) {
+    if (open (COLORDIFFRC, "<$config_file")) {
         while (<COLORDIFFRC>) {
+            my $colourval;
+
             chop;
             next if (/^#/ || /^$/);
             s/\s+//g;
@@ -93,28 +95,43 @@ foreach $config_file (@config_files) {
                 next;
             }
             $setting =~ tr/A-Z/a-z/;
-            $value =~ tr/A-Z/a-z/;
+            $value   =~ tr/A-Z/a-z/;
             if (($value eq 'normal') || ($value eq 'none')) {
                 $value = 'off';
             }
-            if (!defined $colour{$value}) {
+            if ($value =~ m/[0-9]+/ && $value >= 0 && $value <= 255) {
+                # Numeric color
+                if( $value < 8 ) {
+                    $colourval = "\033[0;3${value}m";
+                }
+                elsif( $value < 15 ) {
+                    $colourval = "\033[0;9${value}m";
+                }
+                else {
+                    $colourval = "\033[0;38;5;${value}m";
+                }
+            }
+            elsif (defined($colour{$value})) {
+                $colourval = $colour{$value};
+            }
+            else {
                 print STDERR "Invalid colour specification for setting $setting ($value) in $config_file\n";
                 next;
             }
             if ($setting eq 'plain') {
-                $plain_text = $colour{$value};
+                $plain_text = $colourval;
             }
             elsif ($setting eq 'oldtext') {
-                $file_old = $colour{$value};
+                $file_old = $colourval;
             }
             elsif ($setting eq 'newtext') {
-                $file_new = $colour{$value};
+                $file_new = $colourval;
             }
             elsif ($setting eq 'diffstuff') {
-                $diff_stuff = $colour{$value};
+                $diff_stuff = $colourval;
             }
             elsif ($setting eq 'cvsstuff') {
-                $cvs_stuff = $colour{$value};
+                $cvs_stuff = $colourval;
             }
             else {
                 print STDERR "Unknown option in $config_file: $setting\n";
@@ -127,34 +144,21 @@ foreach $config_file (@config_files) {
 # If output is to a file, switch off colours, unless 'color_patch' is set
 # Relates to http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=378563
 if ((-f STDOUT) && ($color_patch == 0)) {
-    $plain_text='';
-    $file_old='';
-    $file_new='';
-    $diff_stuff='';
-    $cvs_stuff='';
-    $plain_text='';
-    $colour{off}='';
+    $plain_text  = '';
+    $file_old    = '';
+    $file_new    = '';
+    $diff_stuff  = '';
+    $cvs_stuff   = '';
+    $plain_text  = '';
+    $colour{off} = '';
 }
 
+my $specified_difftype;
+GetOptions(
+    "difftype=s" => \$specified_difftype
+);
+# TODO - check that specified type is valid, issue warning if not
 
-# ----------------------------------------------------------------------------
-# FIXME
-#
-# Remove this (contributed) code for now - the long options break the addition
-# of extra diff options supplied on the command line.  On balance, modifying the
-# diff options is probably more important than having colordiff customisations
-# on the command line.
-#
-# Need to find a proper fix for this so that this can be reincluded
-
-###  GetOptions(
-###      "no-banner" => sub { $show_banner = 0 },
-###      "plain-text=s" => \&set_color,
-###      "file-old=s"   => \&set_color,
-###      "file-new=s"   => \&set_color,
-###      "cvs-stuff=s"  => \&set_color,
-###      "diff-stuff=s" => \&set_color
-###  );
 # ----------------------------------------------------------------------------
 
 if ($show_banner == 1) {
@@ -162,19 +166,29 @@ if ($show_banner == 1) {
     print STDERR "$copyright $author, $author_email\n\n";
 }
 
+# FIXME - work out some magic here to make it rip off
+# colordiff-specific long options, such as:
+#
+#   --difftype={plain,context,unified,sideside,debdiff}
+#
+# then, optionally, throw away other options if in a pipe
+
 my @inputstream;
 
-if (defined $ARGV[0]) {
+my $exitcode = 0;
+if ((defined $ARGV[0]) || (-t STDIN)) {
     # More reliable way of pulling in arguments
-    open2(\*INPUTSTREAM, undef, "diff", @ARGV);
+    my $pid = open2(\*INPUTSTREAM, undef, "diff", @ARGV);
     @inputstream = <INPUTSTREAM>;
     close INPUTSTREAM;
+    waitpid $pid, 0;
+    $exitcode=$? >> 8;
 }
 else {
     @inputstream = <STDIN>;
 }
 
-# Input stream has been read - need to examine it 
+# Input stream has been read - need to examine it
 # to determine type of diff we have.
 #
 # This may not be perfect - should identify most reasonably
@@ -185,6 +199,10 @@ my $record;
 my $longest_record = 0;
 
 DIFF_TYPE: foreach $record (@inputstream) {
+    if (defined $specified_difftype) {
+        $diff_type = $specified_difftype;
+        last DIFF_TYPE;
+    }
     # Unified diffs are the only flavour having '+++' or '---'
     # at the start of a line
     if ($record =~ /^(\+\+\+|---|@@)/) {
@@ -231,9 +249,9 @@ my $inside_file_old = 1;
 # second consists only of spaces, '<', '>' and '|'
 # This is not a 100% certain match, but should be good enough
 
-my %separator_col = ();
-my %candidate_col = ();
-my $diffy_sep_col = 0;
+my %separator_col  = ();
+my %candidate_col  = ();
+my $diffy_sep_col  = 0;
 my $mostlikely_sum = 0;
 
 if ($diff_type eq 'diffy') {
@@ -241,34 +259,37 @@ if ($diff_type eq 'diffy') {
     # Unfortunately requires parsing the input stream multiple times
     foreach (@inputstream) {
         # Convert tabs to spaces
-        while ((my $i = index($_, "\t")) > -1) {
-            substr(
-                    $_, $i, 1,                # range to replace
-                    (' ' x (8 - ($i % 8))),     # string to replace with
-                  );
+        while ((my $i = index ($_, "\t")) > -1) {
+            substr (
+                $_, $i, 1,    # range to replace
+                (' ' x (8 - ($i % 8))),    # string to replace with
+            );
         }
         $record = $_;
-        $longest_record = length($record) if (length($record) > $longest_record);
+        $longest_record = length ($record) if (length ($record) > $longest_record);
     }
-    for(my $i=0;$i <= $longest_record;$i++) {
+    for (my $i = 0 ; $i <= $longest_record ; $i++) {
         $separator_col{$i} = 1;
         $candidate_col{$i} = 0;
     }
 
     foreach (@inputstream) {
         # Convert tabs to spaces
-        while ((my $i = index($_, "\t")) > -1) {
-            substr(
-                    $_, $i, 1,                # range to replace
-                    (' ' x (8 - ($i % 8))),     # string to replace with
-                  );
+        while ((my $i = index ($_, "\t")) > -1) {
+            substr (
+                $_, $i, 1,    # range to replace
+                (' ' x (8 - ($i % 8))),    # string to replace with
+            );
         }
-        for(my $i=0;$i<(length($_)-2);$i++) {
+        for (my $i = 0 ; $i < (length ($_) - 2) ; $i++) {
             next if (!defined $separator_col{$i});
             next if ($separator_col{$i} == 0);
-            my $subsub = substr($_,$i,2);
-            if ( ($subsub ne '  ') && ($subsub ne ' |') && ($subsub ne ' >') &&
-                    ($subsub ne ' <') ) {
+            my $subsub = substr ($_, $i, 2);
+            if (   ($subsub ne '  ')
+                && ($subsub ne ' |')
+                && ($subsub ne ' >')
+                && ($subsub ne ' <'))
+            {
                 $separator_col{$i} = 0;
             }
             if (($subsub eq ' |') || ($subsub eq ' >') || ($subsub eq ' <')) {
@@ -277,10 +298,10 @@ if ($diff_type eq 'diffy') {
         }
     }
 
-    for(my $i=0;$i<$longest_record-2;$i++) {
+    for (my $i = 0 ; $i < $longest_record - 2 ; $i++) {
         if ($separator_col{$i} == 1) {
             if ($candidate_col{$i} > $mostlikely_sum) {
-                $diffy_sep_col = $i;
+                $diffy_sep_col  = $i;
                 $mostlikely_sum = $i;
             }
         }
@@ -302,6 +323,9 @@ foreach (@inputstream) {
         elsif (/^(Index: |={4,}|RCS file: |retrieving |diff )/) {
             print "$cvs_stuff";
         }
+        elsif (/^Only in/) {
+            print "$diff_stuff";
+        }
         else {
             print "$plain_text";
         }
@@ -316,6 +340,9 @@ foreach (@inputstream) {
         elsif (/^\*{4,}/) {
             print "$diff_stuff";
         }
+        elsif (/^Only in/) {
+            print "$diff_stuff";
+        }
         elsif (/^\*\*\* [0-9]+,[0-9]+/) {
             print "$diff_stuff";
             $inside_file_old = 1;
@@ -355,6 +382,9 @@ foreach (@inputstream) {
         elsif (/^\@/) {
             print "$diff_stuff";
         }
+        elsif (/^Only in/) {
+            print "$diff_stuff";
+        }
         elsif (/^(Index: |={4,}|RCS file: |retrieving |diff )/) {
             print "$cvs_stuff";
         }
@@ -365,8 +395,8 @@ foreach (@inputstream) {
     # Works with previously-identified column containing the diff-y
     # separator characters
     elsif ($diff_type eq 'diffy') {
-        if (length($_) > ($diffy_sep_col+2)) {
-            my $sepchars = substr($_, $diffy_sep_col, 2);
+        if (length ($_) > ($diffy_sep_col + 2)) {
+            my $sepchars = substr ($_, $diffy_sep_col, 2);
             if ($sepchars eq ' <') {
                 print "$file_old";
             }
@@ -380,6 +410,9 @@ foreach (@inputstream) {
                 print "$plain_text";
             }
         }
+        elsif (/^Only in/) {
+            print "$diff_stuff";
+        }
         else {
             print "$plain_text";
         }
@@ -388,14 +421,12 @@ foreach (@inputstream) {
         $_ =~ s/(\[-[^]]*?-\])/$file_old$1$colour{off}/g;
         $_ =~ s/(\{\+[^]]*?\+\})/$file_new$1$colour{off}/g;
     }
+    elsif ($diff_type eq 'debdiff') {
+        $_ =~ s/(\[-[^]]*?-\])/$file_old$1$colour{off}/g;
+        $_ =~ s/(\{\+[^]]*?\+\})/$file_new$1$colour{off}/g;
+    }
     s/$/$colour{off}/;
     print "$_";
 }
 
-#   FIXME - See earlier commented-out code
-### sub set_color {
-###     my ($type, $color) = @_;
-### 
-###     $type =~ s/-/_/;
-###     eval "\$$type = \$colour{$color}";
-### }
+exit $exitcode;
